@@ -1,39 +1,57 @@
 import abc
 from copy import copy
 from functools import partial, singledispatch
-from typing import Callable, Dict, List, Optional, Tuple, Union
+from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Tuple, Union
 
 import aesara.tensor as at
 from aesara.gradient import DisconnectedType, jacobian
-from aesara.graph.basic import Apply, Node, Variable
+from aesara.graph.basic import Apply, Variable, walk
 from aesara.graph.features import AlreadyThere, Feature
 from aesara.graph.fg import FunctionGraph
 from aesara.graph.op import Op
 from aesara.graph.rewriting.basic import GraphRewriter, in2out, node_rewriter
-from aesara.scalar import Add, Exp, Log, Mul
-from aesara.tensor.elemwise import Elemwise
-from aesara.tensor.rewriting.basic import (
-    register_specialize,
-    register_stabilize,
-    register_useless,
-)
+from aesara.tensor.math import add, exp, log, mul, reciprocal, sub, true_div
+from aesara.tensor.rewriting.basic import register_useless
 from aesara.tensor.var import TensorVariable
+from typing_extensions import Protocol
 
 from aeppl.abstract import (
     MeasurableElemwise,
     MeasurableVariable,
+    ValuedVariable,
     _get_measurable_outputs,
     assign_custom_measurable_outputs,
+    valued_variable,
 )
 from aeppl.logprob import _logprob, logprob
-from aeppl.rewriting import PreserveRVMappings, measurable_ir_rewrites_db
-from aeppl.utils import walk_model
+from aeppl.rewriting import measurable_ir_rewrites_db
+
+if TYPE_CHECKING:
+    from aesara.graph.rewriting.basic import NodeRewriter
+
+    class TransformFnType(Protocol):
+        def __call__(
+            self, measurable_input: MeasurableVariable, *other_inputs: Variable
+        ) -> Tuple["RVTransform", Tuple[TensorVariable, ...]]:
+            pass
+
+
+def register_measurable_ir(
+    node_rewriter: "NodeRewriter",
+    *tags: str,
+    **kwargs,
+):
+    name = kwargs.pop("name", None) or node_rewriter.__name__
+    measurable_ir_rewrites_db.register(
+        name, node_rewriter, "basic", "transform", *tags, **kwargs
+    )
+    return node_rewriter
 
 
 @singledispatch
 def _default_transformed_rv(
     op: Op,
-    node: Node,
+    node: Apply,
 ) -> Optional[Apply]:
     """Create a node for a transformed log-probability of a `MeasurableVariable`.
 
@@ -46,7 +64,16 @@ def _default_transformed_rv(
 
 
 class TransformedVariable(Op):
-    """A no-op that identifies a transform and its un-transformed input."""
+    r"""A no-op that identifies a transformed value variable and its un-transformed input.
+
+    This transformed/untransformed-value pairing is primarily used as a means to
+    obtain the untransformed value for use with the `RVTransform.log_jac_det` when
+    log-probabilities are generated.
+
+    It is expected that all occurrences of these `Op`\s will be removed during
+    compilation by `remove_TransformedVariables`.
+
+    """
 
     view_map = {0: [0]}
 
@@ -71,8 +98,6 @@ class TransformedVariable(Op):
 transformed_variable = TransformedVariable()
 
 
-@register_specialize
-@register_stabilize
 @register_useless
 @node_rewriter([TransformedVariable])
 def remove_TransformedVariables(fgraph, node):
@@ -108,9 +133,9 @@ class DefaultTransformSentinel:
 DEFAULT_TRANSFORM = DefaultTransformSentinel()
 
 
-@node_rewriter(tracks=None)
-def transform_values(fgraph: FunctionGraph, node: Node) -> Optional[List[Node]]:
-    """Apply transforms to value variables.
+@node_rewriter([ValuedVariable])
+def transform_values(fgraph: FunctionGraph, node: Apply):
+    r"""Apply transforms to the values of value-bound measurable variables.
 
     It is assumed that the input value variables correspond to forward
     transformations, usually chosen in such a way that the values are
@@ -119,22 +144,48 @@ def transform_values(fgraph: FunctionGraph, node: Node) -> Optional[List[Node]]:
     For example, if ``Y = halfnormal(...)``, we assume the respective value
     variable is specified on the log scale and back-transform it to obtain
     ``Y`` on the natural scale.
+
+    The main steps of this rewrite are as follows:
+
+    1. Obtain a `RVTransform` for a `ValuedVariable`.
+    2. Replace all occurrences of the original value variable with the
+    "backward"-transform of a new value variable in the "forward"-transformed
+    space.
+    3. Signify that the original value variable is to be replaced by the new
+    one.
+    4. Replace the old `ValuedVariable` with a new one containing a
+    `TransformedVariable` value.
+
+    Step 3. is currently accomplished by updating the `memo` dictionary
+    associated with the `FunctionGraph`.  Our main entry-point,
+    `conditional_logprob`, checks this dictionary for value variable changes.
+
+    TODO: This approach is less than ideal, because it puts awkward demands on
+    users/callers of this rewrite to check with `memo`; let's see if we can do
+    something better.
+
+    The new value variable mentioned in Step 2. may be of a different `Type`
+    (e.g. extra/fewer dimensions) than the original value variable; this is why
+    we must replace the corresponding original value variables before we
+    construct log-probability graphs.  This is due to the fact that we
+    generally cannot replace variables with new ones that have different
+    `Type`\s.
+
     """
 
-    rv_map_feature = getattr(fgraph, "preserve_rv_mappings", None)
     values_to_transforms = getattr(fgraph, "values_to_transforms", None)
 
-    if rv_map_feature is None or values_to_transforms is None:
+    if values_to_transforms is None:
         return None  # pragma: no cover
 
-    try:
-        rv_var = node.default_output()
-        rv_var_out_idx = node.outputs.index(rv_var)
-    except ValueError:
-        return None
+    rv_var, value_var = node.inputs
 
-    value_var = rv_map_feature.rv_values.get(rv_var, None)
-    if value_var is None:
+    rv_node = rv_var.owner
+
+    try:
+        rv_var = rv_node.default_output()
+        rv_var_out_idx = rv_node.outputs.index(rv_var)
+    except ValueError:
         return None
 
     transform = values_to_transforms.get(value_var, None)
@@ -142,82 +193,154 @@ def transform_values(fgraph: FunctionGraph, node: Node) -> Optional[List[Node]]:
     if transform is None:
         return None
     elif transform is DEFAULT_TRANSFORM:
-        trans_node = _default_transformed_rv(node.op, node)
+        trans_node = _default_transformed_rv(rv_node.op, rv_node)
         if trans_node is None:
             return None
         transform = trans_node.op.transform
     else:
-        new_op = _create_transformed_rv_op(node.op, transform)
+        # This automatically constructs a `_logprob` dispatch for the
+        # transformed node, so look there for the log-probability
+        # implementation
+        new_op = _create_transformed_rv_op(rv_node.op, transform)
         # Create a new `Apply` node and outputs
-        trans_node = node.clone()
+        trans_node = rv_node.clone()
         trans_node.op = new_op
-        trans_node.outputs[rv_var_out_idx].name = node.outputs[rv_var_out_idx].name
+        trans_node.outputs[rv_var_out_idx].name = rv_node.outputs[rv_var_out_idx].name
 
     # We now assume that the old value variable represents the *transformed space*.
-    # This means that we need to replace all instance of the old value variable
-    # with "inversely/un-" transformed versions of itself.
-    new_value_var = transformed_variable(
-        transform.backward(value_var, *trans_node.inputs), value_var
-    )
-    if value_var.name and getattr(transform, "name", None):
-        new_value_var.name = f"{value_var.name}_{transform.name}"
 
-    rv_map_feature.update_rv_maps(
-        rv_var, new_value_var, trans_node.outputs[rv_var_out_idx]
+    # We normally initialize value variables as copies of the random variables,
+    # thus, in the untransformed space, we need to apply the forward
+    # transformation to get value variables with the correct `Type`s in the
+    # transformed space.
+    trans_value_var: TensorVariable = transform.forward(
+        value_var, *trans_node.inputs
+    ).type()
+
+    if value_var.name:
+        trans_value_var.name = f"{value_var.name}-trans"
+
+    fgraph.add_input(trans_value_var)
+
+    # We need to replace all instances of the old value variables
+    # with "inversely/un-" transformed versions of themselves.
+    untrans_value_var = transformed_variable(
+        transform.backward(trans_value_var, *trans_node.inputs), trans_value_var
     )
 
-    return trans_node.outputs
+    # This effectively lets the caller know that a value variable has been
+    # replaced (i.e. they should filter all their old value variables through
+    # the memo/replacements map).
+    fgraph.memo[value_var] = trans_value_var
+
+    trans_var = trans_node.outputs[rv_var_out_idx]
+    new_var = valued_variable(trans_var, untrans_value_var)
+
+    return {value_var: untrans_value_var, node.outputs[0]: new_var}
 
 
 class TransformValuesMapping(Feature):
-    r"""A `Feature` that maintains a map between value variables and their transforms."""
+    r"""A `Feature` that maintains a map between value variables and their transforms as
+    well as between value variables and their transformed counterparts.
 
-    def __init__(self, values_to_transforms):
+    This is exclusively/primarily used by `TransformValuesRewrite`.
+
+    """
+
+    def __init__(self, values_to_transforms, memo):
+        """
+        Parameters
+        ==========
+        values_to_transforms
+            Mapping between value variables and their transformations. Each
+            value variable can be assigned one of `RVTransform`,
+            `DEFAULT_TRANSFORM`, or ``None``. Random variables with no
+            transform specified remain unchanged.
+        memo
+            Mapping from variables to their clones.  This is updated
+            in-place whenever a value variable is transformed.
+
+        """
         self.values_to_transforms = values_to_transforms
+        self.memo = memo
 
     def on_attach(self, fgraph):
         if hasattr(fgraph, "values_to_transforms"):
             raise AlreadyThere()
 
         fgraph.values_to_transforms = self.values_to_transforms
+        fgraph.memo = self.memo
 
 
 class TransformValuesRewrite(GraphRewriter):
-    r"""Transforms value variables according to a map and/or per-`RandomVariable` defaults."""
+    r"""Transforms measurable variables.
+
+    This transformation can be used to replace all occurrences of a measurable
+    variable and its values within a model graph with a transformed version of
+    itself (e.g. for the purposes of sampling on "unconstrained" ranges).
+
+    Example
+    =======
+
+    .. code::
+
+        A_rv = at.random.normal(0, 1.0, name="A")
+        B_rv = at.random.normal(A_rv, 1.0, name="B")
+
+        logp, (a_vv, b_vv) = joint_logprob(
+            A_rv,
+            B_rv,
+            extra_rewrites=TransformValuesRewrite(
+                {A_rv: ExpTransform()}
+            ),
+        )
+
+        >>> print(aesara.pprint(logp))
+        sum([((-0.9189385332046727 + (-0.5 * (log(A_vv-trans) ** 2))) + (-1.0 * log(A_vv-trans))),
+             (-0.9189385332046727 + (-0.5 * ((B_vv - log(A_vv-trans)) ** 2)))], axis=None)
+
+    """
 
     default_transform_rewrite = in2out(transform_values, ignore_newtrees=True)
 
     def __init__(
         self,
-        values_to_transforms: Dict[
+        rvs_to_transforms: Dict[
             TensorVariable, Union[RVTransform, DefaultTransformSentinel, None]
         ],
     ):
         """
         Parameters
         ==========
-        values_to_transforms
-            Mapping between value variables and their transformations.  Each
-            value variable can be assigned one of `RVTransform`,
-            ``DEFAULT_TRANSFORM``, or ``None``. If a transform is not specified
-            for a specific value variable it will not be transformed.
+        rvs_to_transforms
+            Mapping between measurable variables and their transformations. Each
+            measurable variable can be assigned an `RVTransform` instance,
+            `DEFAULT_TRANSFORM`, or ``None``. Measurable variables with no
+            transform specified remain unchanged.
 
         """
 
-        self.values_to_transforms = values_to_transforms
+        self.rvs_to_transforms = rvs_to_transforms
 
-    def add_requirements(self, fgraph):
-        values_transforms_feature = TransformValuesMapping(self.values_to_transforms)
+    def add_requirements(
+        self,
+        fgraph,
+        rv_to_values: Dict[TensorVariable, TensorVariable],
+        memo: Dict[TensorVariable, TensorVariable],
+    ):
+        values_to_transforms = {
+            rv_to_values[rv]: transform
+            for rv, transform in self.rvs_to_transforms.items()
+        }
+        values_transforms_feature = TransformValuesMapping(values_to_transforms, memo)
         fgraph.attach_feature(values_transforms_feature)
 
     def apply(self, fgraph: FunctionGraph):
         return self.default_transform_rewrite.rewrite(fgraph)
 
 
-class MeasurableTransform(MeasurableElemwise):
-    """A placeholder used to specify a log-likelihood for a transformed measurable variable"""
-
-    valid_scalar_types = (Exp, Log, Add, Mul)
+class MeasurableElemwiseTransform(MeasurableElemwise):
+    """A placeholder used to specify a log-likelihood for a transformed `Elemwise`."""
 
     # Cannot use `transform` as name because it would clash with the property added by
     # the `TransformValuesRewrite`
@@ -232,14 +355,16 @@ class MeasurableTransform(MeasurableElemwise):
         super().__init__(*args, **kwargs)
 
 
-@_get_measurable_outputs.register(MeasurableTransform)
-def _get_measurable_outputs_Transform(op, node):
+@_get_measurable_outputs.register(MeasurableElemwiseTransform)
+def _get_measurable_outputs_ElemwiseTransform(op, node):
     return [node.default_output()]
 
 
-@_logprob.register(MeasurableTransform)
-def measurable_transform_logprob(op: MeasurableTransform, values, *inputs, **kwargs):
-    """Compute the log-probability graph for a `MeasurabeTransform`."""
+@_logprob.register(MeasurableElemwiseTransform)
+def measurable_elemwise_logprob(
+    op: MeasurableElemwiseTransform, values, *inputs, **kwargs
+):
+    """Compute the log-probability graph for a `MeasurableElemwiseTransform`."""
     # TODO: Could other rewrites affect the order of inputs?
     (value,) = values
     other_inputs = list(inputs)
@@ -255,22 +380,151 @@ def measurable_transform_logprob(op: MeasurableTransform, values, *inputs, **kwa
     return input_logprob + jacobian
 
 
-@node_rewriter([Elemwise])
-def find_measurable_transforms(
-    fgraph: FunctionGraph, node: Node
-) -> Optional[List[Node]]:
-    """Find measurable transformations from Elemwise operators."""
+@register_measurable_ir
+@node_rewriter([true_div])
+def measurable_true_div(fgraph, node):
+    r"""Rewrite a `true_div` node to a `MeasurableVariable`.
+
+    TODO FIXME: We need update/clarify the canonicalization situation so that
+    these can be reliably rewritten as products of reciprocals.
+
+    """
+    numerator, denominator = node.inputs
+
+    reciprocal_denominator = at.reciprocal(denominator)
+    # `denominator` is measurable
+    res = measurable_reciprocal.transform(fgraph, reciprocal_denominator.owner)
+    if res:
+        return measurable_mul.transform(fgraph, at.mul(numerator, res[0]).owner)
+
+    # `numerator` is measurable
+    return measurable_mul.transform(
+        fgraph, at.mul(numerator, reciprocal_denominator).owner
+    )
+
+
+@register_measurable_ir
+@node_rewriter([sub])
+def measurable_sub(fgraph, node):
+    r"""Rewrite a `sub` node to a `MeasurableVariable`.
+
+    TODO FIXME: We need update/clarify the canonicalization situation so that
+    these can be reliably rewritten as products of reciprocals.
+
+    """
+    minuend, subtrahend = node.inputs
+
+    mul_subtrahend = at.mul(-1, subtrahend)
+
+    # `subtrahend` is measurable
+    res = measurable_mul.transform(fgraph, mul_subtrahend.owner)
+    if res:
+        return measurable_add.transform(fgraph, at.add(minuend, res[0]).owner)
+
+    # TODO FIXME: `local_add_canonizer` will unreliably rewrite expressions like
+    # `x - y` to `-y + x` (e.g. apparently when `y` is a constant?) and, as a result,
+    # this will not be reached.  We're leaving this in just in case, but we
+    # ultimately need to fix Aesara's canonicalizations.
+
+    # `minuend` is measurable
+    return measurable_add.transform(fgraph, at.add(minuend, mul_subtrahend).owner)
+
+
+@register_measurable_ir
+@node_rewriter([exp])
+def measurable_exp(fgraph, node):
+    """Rewrite an `exp` node to a `MeasurableVariable`."""
+
+    def transform(measurable_input, *args):
+        return ExpTransform(), (measurable_input,)
+
+    return construct_elemwise_transform(fgraph, node, transform)
+
+
+@register_measurable_ir
+@node_rewriter([log])
+def measurable_log(fgraph, node):
+    """Rewrite a `log` node to a `MeasurableVariable`."""
+
+    def transform(measurable_input, *args):
+        return LogTransform(), (measurable_input,)
+
+    return construct_elemwise_transform(fgraph, node, transform)
+
+
+@register_measurable_ir
+@node_rewriter([add])
+def measurable_add(fgraph, node):
+    """Rewrite an `add` node to a `MeasurableVariable`."""
+
+    def transform(measurable_input, *other_inputs):
+        transform_inputs = (
+            measurable_input,
+            at.add(*other_inputs) if len(other_inputs) > 1 else other_inputs[0],
+        )
+        transform = LocTransform(
+            transform_args_fn=lambda *inputs: inputs[-1],
+        )
+        return transform, transform_inputs
+
+    return construct_elemwise_transform(fgraph, node, transform)
+
+
+@register_measurable_ir
+@node_rewriter([mul])
+def measurable_mul(fgraph, node):
+    """Rewrite a `mul` node to a `MeasurableVariable`."""
+
+    def transform(measurable_input, *other_inputs):
+        transform_inputs = (
+            measurable_input,
+            at.mul(*other_inputs) if len(other_inputs) > 1 else other_inputs[0],
+        )
+        return (
+            ScaleTransform(
+                transform_args_fn=lambda *inputs: inputs[-1],
+            ),
+            transform_inputs,
+        )
+
+    return construct_elemwise_transform(fgraph, node, transform)
+
+
+@register_measurable_ir
+@node_rewriter([reciprocal])
+def measurable_reciprocal(fgraph, node):
+    """Rewrite a `reciprocal` node to a `MeasurableVariable`."""
+
+    def transform(measurable_input, *other_inputs):
+        return ReciprocalTransform(), (measurable_input,)
+
+    return construct_elemwise_transform(fgraph, node, transform)
+
+
+def construct_elemwise_transform(
+    fgraph: FunctionGraph,
+    node: Apply,
+    transform_fn: "TransformFnType",
+) -> Optional[List[Variable]]:
+    """Construct a measurable transformation for an `Elemwise` node.
+
+    Parameters
+    ----------
+    fgraph
+        The `FunctionGraph` in which `node` resides.
+    node
+        The `Apply` node to be converted.
+    transform_fn
+        A function that takes a single measurable input and all the remaining
+        inputs and returns a transform object and transformed inputs.
+
+    Returns
+    -------
+    A new variable with an `Apply` node with a `MeasurableElemwiseTransform`
+    that replaces `node`.
+
+    """
     scalar_op = node.op.scalar_op
-    if not isinstance(scalar_op, MeasurableTransform.valid_scalar_types):
-        return None
-
-    # Node was already converted
-    if isinstance(node.op, MeasurableVariable):
-        return None  # pragma: no cover
-
-    rv_map_feature: PreserveRVMappings = getattr(fgraph, "preserve_rv_mappings", None)
-    if rv_map_feature is None:
-        return None  # pragma: no cover
 
     # Check that we have a single source of measurement
     measurable_inputs = [
@@ -278,7 +532,7 @@ def find_measurable_transforms(
         for idx, inp in enumerate(node.inputs)
         if inp.owner
         and isinstance(inp.owner.op, MeasurableVariable)
-        and inp not in rv_map_feature.rv_values
+        and not isinstance(inp, ValuedVariable)
     ]
 
     if len(measurable_inputs) != 1:
@@ -287,50 +541,49 @@ def find_measurable_transforms(
     measurable_input: TensorVariable = measurable_inputs[0]
 
     # Do not apply rewrite to discrete variables
+    # TODO: Formalize this restriction better.
     if measurable_input.type.dtype.startswith("int"):
         return None
 
     # Check that other inputs are not potentially measurable, in which case this rewrite
     # would be invalid
+    # TODO FIXME: This is rather costly and redundant; find a way to avoid it
+    # or make it cheaper.
     other_inputs = tuple(inp for inp in node.inputs if inp is not measurable_input)
+
+    def expand(var: TensorVariable) -> List[TensorVariable]:
+        new_vars: List[TensorVariable] = []
+
+        if (
+            var.owner
+            and not isinstance(var.owner.op, MeasurableVariable)
+            and not isinstance(var, ValuedVariable)
+        ):
+            new_vars.extend(reversed(var.owner.inputs))
+
+        return new_vars
+
     if any(
         ancestor_node
-        for ancestor_node in walk_model(
-            other_inputs,
-            walk_past_rvs=False,
-            stop_at_vars=set(rv_map_feature.rv_values),
-        )
+        for ancestor_node in walk(other_inputs, expand, False)
         if (
             ancestor_node.owner
             and isinstance(ancestor_node.owner.op, MeasurableVariable)
-            and ancestor_node not in rv_map_feature.rv_values
+            and not isinstance(ancestor_node, ValuedVariable)
         )
     ):
         return None
 
     # Make base_measure outputs unmeasurable
     # This seems to be the only thing preventing nested rewrites from being erased
-    measurable_input = assign_custom_measurable_outputs(measurable_input.owner)
+    # TODO: Is this still needed?
+    if not isinstance(measurable_input.owner.op, ValuedVariable):
+        measurable_input = assign_custom_measurable_outputs(measurable_input.owner)
 
     measurable_input_idx = 0
-    transform_inputs: Tuple[TensorVariable, ...] = (measurable_input,)
-    transform: RVTransform
-    if isinstance(scalar_op, Exp):
-        transform = ExpTransform()
-    elif isinstance(scalar_op, Log):
-        transform = LogTransform()
-    elif isinstance(scalar_op, Add):
-        transform_inputs = (measurable_input, at.add(*other_inputs))
-        transform = LocTransform(
-            transform_args_fn=lambda *inputs: inputs[-1],
-        )
-    else:
-        transform_inputs = (measurable_input, at.mul(*other_inputs))
-        transform = ScaleTransform(
-            transform_args_fn=lambda *inputs: inputs[-1],
-        )
+    transform, transform_inputs = transform_fn(measurable_input, *other_inputs)
 
-    transform_op = MeasurableTransform(
+    transform_op = MeasurableElemwiseTransform(
         scalar_op=scalar_op,
         transform=transform,
         measurable_input_idx=measurable_input_idx,
@@ -341,15 +594,6 @@ def find_measurable_transforms(
     return [transform_out]
 
 
-measurable_ir_rewrites_db.register(
-    "find_measurable_transforms",
-    find_measurable_transforms,
-    0,
-    "basic",
-    "transform",
-)
-
-
 class LocTransform(RVTransform):
     name = "loc"
 
@@ -413,6 +657,19 @@ class ExpTransform(RVTransform):
         return -at.log(value)
 
 
+class ReciprocalTransform(RVTransform):
+    name = "reciprocal"
+
+    def forward(self, value, *inputs):
+        return at.reciprocal(value)
+
+    def backward(self, value, *inputs):
+        return at.reciprocal(value)
+
+    def log_jac_det(self, value, *inputs):
+        return -2 * at.log(value)
+
+
 class IntervalTransform(RVTransform):
     name = "interval"
 
@@ -608,12 +865,14 @@ def _create_transformed_rv_op(
         """
         (value,) = values
 
-        logprob = _logprob(rv_op, values, *inputs, **kwargs)
+        assert value.owner.op == transformed_variable
+
+        backward_val, trans_val = value.owner.inputs
+
+        logprob = _logprob(rv_op, (backward_val,), *inputs, **kwargs)
 
         if use_jacobian:
-            assert isinstance(value.owner.op, TransformedVariable)
-            original_forward_value = value.owner.inputs[1]
-            jacobian = op.transform.log_jac_det(original_forward_value, *inputs)
+            jacobian = op.transform.log_jac_det(trans_val, *inputs)
             logprob += jacobian
 
         return logprob
